"""
    Copyright 2024 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    schema = """
    UPDATE public.configurationmodel
    SET undeployable=(
        CASE
            WHEN undeployable IS NULL
            -- Set default to empty array
            THEN '{}'
            -- Keep current value
            ELSE undeployable
        END
    ),
    skipped_for_undeployable=(
        CASE
            WHEN skipped_for_undeployable IS NULL
            -- Set default to empty array
            THEN '{}'
            -- Keep current value
            ELSE skipped_for_undeployable
        END
    );

     -- change the type of the 'undeployable' and 'skipped_for_undeployable' columns to disallow null values

     ALTER TABLE public.configurationmodel
         ALTER COLUMN undeployable SET NOT NULL,
         ALTER COLUMN skipped_for_undeployable SET NOT NULL;
     """

    await connection.execute(schema)
